@@ -5,8 +5,10 @@ import os
 import time
 import subprocess
 import random
-import traceback
 import queue
+import logging
+
+log = logging.getLogger('mpv-jsonipc')
 
 if os.name == "nt":
     import _winapi
@@ -14,6 +16,20 @@ if os.name == "nt":
 
 TIMEOUT = 120
 
+# Older MPV versions do not allow us to dynamically retrieve the command list.
+FALLBACK_COMMAND_LIST = [
+    'ignore', 'seek', 'revert-seek', 'quit', 'quit-watch-later', 'stop', 'frame-step', 'frame-back-step',
+    'playlist-next', 'playlist-prev', 'playlist-shuffle', 'playlist-unshuffle', 'sub-step', 'sub-seek',
+    'print-text', 'show-text', 'expand-text', 'expand-path', 'show-progress', 'sub-add', 'audio-add',
+    'video-add', 'sub-remove', 'audio-remove', 'video-remove', 'sub-reload', 'audio-reload', 'video-reload',
+    'rescan-external-files', 'screenshot', 'screenshot-to-file', 'screenshot-raw', 'loadfile', 'loadlist',
+    'playlist-clear', 'playlist-remove', 'playlist-move', 'run', 'subprocess', 'set', 'change-list', 'add',
+    'cycle', 'multiply', 'cycle-values', 'enable-section', 'disable-section', 'define-section', 'ab-loop',
+    'drop-buffers', 'af', 'vf', 'af-command', 'vf-command', 'ao-reload', 'script-binding', 'script-message',
+    'script-message-to', 'overlay-add', 'overlay-remove', 'osd-overlay', 'write-watch-later-config',
+    'hook-add', 'hook-ack', 'mouse', 'keybind', 'keypress', 'keydown', 'keyup', 'apply-profile',
+    'load-script', 'dump-cache', 'ab-loop-dump-cache', 'ab-loop-align-cache']
+
 class MPVError(Exception):
     def __init__(self, *args, **kwargs):
         super(MPVError, self).__init__(*args, **kwargs)
@@ -23,10 +39,9 @@ class WindowsSocket(threading.Thread):
         ipc_socket = "\\\\.\\pipe\\" + ipc_socket
         self.callback = callback
         
-        openmode = _winapi.PIPE_ACCESS_DUPLEX
         access = _winapi.GENERIC_READ | _winapi.GENERIC_WRITE
         limit = 5 # Connection may fail at first. Try 5 times.
-        for _ in range(5):
+        for _ in range(limit):
             try:
                 pipe_handle = _winapi.CreateFile(
                     ipc_socket, access, 0, _winapi.NULL, _winapi.OPEN_EXISTING,
@@ -122,6 +137,7 @@ class MPVProcess:
             else:
                 mpv_location = "mpv"
         
+        log.debug("Staring MPV from {0}.".format(mpv_location))
         ipc_socket_name = ipc_socket
         if os.name == 'nt':
             ipc_socket = "\\\\.\\pipe\\" + ipc_socket
@@ -129,6 +145,7 @@ class MPVProcess:
         if os.name != 'nt' and os.path.exists(ipc_socket):
             os.remove(ipc_socket)
 
+        log.debug("Using IPC socket {0} for MPV.".format(ipc_socket))
         self.ipc_socket = ipc_socket
         args = [mpv_location]
         self._set_default(kwargs, "idle", True)
@@ -138,15 +155,14 @@ class MPVProcess:
         args.extend("--{0}={1}".format(v[0].replace("_", "-"), self._mpv_fmt(v[1]))
                     for v in kwargs.items())
         self.process = subprocess.Popen(args)
-        time.sleep(1)
-        for _ in range(20):
+        for _ in range(100): # Give MPV 10 seconds to start.
             time.sleep(0.1)
             self.process.poll()
             if os.path.exists(ipc_socket) or self.process.returncode is not None:
                 break
         
         if not os.path.exists(ipc_socket) or self.process.returncode is not None:
-            print("dead", os.path.exists(ipc_socket), self.process.returncode)
+            self.process.terminate()
             raise MPVError("MPV not started.")
 
     def _set_default(self, prop_dict, key, value):
@@ -244,8 +260,8 @@ class EventHandler(threading.Thread):
                 break
             try:
                 event[0](*event[1])
-            except:
-                traceback.print_exc()
+            except Exception:
+                log.error("EventHandler caught exception from {0}.".format(event), exc_info=1)
 
 class MPV:
     def __init__(self, start_mpv=True, ipc_socket=None, mpv_location=None, **kwargs):
@@ -264,12 +280,26 @@ class MPV:
                 ipc_socket = "/tmp/{0}".format(rand_file)
 
         if start_mpv:
-            self.mpv_process = MPVProcess(ipc_socket, mpv_location, **kwargs)
+            # Attempt to start MPV 3 times.
+            for i in range(3):
+                try:
+                    self.mpv_process = MPVProcess(ipc_socket, mpv_location, **kwargs)
+                    break
+                except MPVError:
+                    log.warning("MPV start failed.", exc_info=1)
+                    continue
+            else:
+                raise MPVError("MPV process retry limit reached.")
 
         self.mpv_inter = MPVInter(ipc_socket, self._callback)
         self.properties = set(x.replace("-", "_") for x in self.command("get_property", "property-list"))
-        for command in self.command("get_property", "command-list"):
-            object.__setattr__(self, command["name"].replace("-", "_"), self._get_wrapper(command["name"]))
+        try:
+            command_list = [x["name"] for x in self.command("get_property", "command-list")]
+        except MPVError:
+            log.warning("Using fallback command list.")
+            command_list = FALLBACK_COMMAND_LIST
+        for command in command_list:
+            object.__setattr__(self, command.replace("-", "_"), self._get_wrapper(command))
 
         self._dir = list(self.properties)
         self._dir.extend(object.__dir__(self))
@@ -301,6 +331,10 @@ class MPV:
             return func
         return wrapper
 
+    # Added for compatibility.
+    def event_callback(self, name):
+        return self.on_event(name)
+
     def on_key_press(self, name):
         def wrapper(func):
             self.bind_key_press(name, func)
@@ -313,8 +347,13 @@ class MPV:
         self.keybind_id += 1
         self.keybind_lock.release()
 
+        bind_name = "bind{0}".format(keybind_id)
         self.key_bindings["bind{0}".format(keybind_id)] = callback
-        self.keybind(name, "script-message custom-bind bind{0}".format(keybind_id))
+        try:
+            self.keybind(name, "script-message custom-bind {0}".format(bind_name))
+        except MPVError:
+            self.define_section(bind_name, "{0} script-message custom-bind {1}".format(name, bind_name))
+            self.enable_section(bind_name)
 
     def bind_property_observer(self, name, callback):
         self.observer_lock.acquire()
